import abc
from typing import (
    Generic,
    Iterable,
    Iterator,
    Mapping,
    MutableMapping,
    TypeVar,
    overload,
)

class istr(str): ...

upstr = istr

_S = str | istr

_T = TypeVar("_T")

_T_co = TypeVar("_T_co", covariant=True)

_D = TypeVar("_D")

class MultiMapping(Mapping[_S, _T_co]):
    @overload
    @abc.abstractmethod
    def getall(self, key: _S) -> list[_T_co]: ...
    @overload
    @abc.abstractmethod
    def getall(self, key: _S, default: _D) -> list[_T_co] | _D: ...
    @overload
    @abc.abstractmethod
    def getone(self, key: _S) -> _T_co: ...
    @overload
    @abc.abstractmethod
    def getone(self, key: _S, default: _D) -> _T_co | _D: ...

_Arg = (Mapping[str, _T] | Mapping[istr, _T] | dict[str, _T]
        | dict[istr, _T] | MultiMapping[_T]
        | Iterable[tuple[str, _T]] | Iterable[tuple[istr, _T]])

class MutableMultiMapping(MultiMapping[_T], MutableMapping[_S, _T], Generic[_T]):
    @abc.abstractmethod
    def add(self, key: _S, value: _T) -> None: ...
    @abc.abstractmethod
    def extend(self, arg: _Arg[_T] = ..., **kwargs: _T) -> None: ...
    @overload
    @abc.abstractmethod
    def popone(self, key: _S) -> _T: ...
    @overload
    @abc.abstractmethod
    def popone(self, key: _S, default: _D) -> _T | _D: ...
    @overload
    @abc.abstractmethod
    def popall(self, key: _S) -> list[_T]: ...
    @overload
    @abc.abstractmethod
    def popall(self, key: _S, default: _D) -> list[_T] | _D: ...

class MultiDict(MutableMultiMapping[_T], Generic[_T]):
    def __init__(self, arg: _Arg[_T] = ..., **kwargs: _T) -> None: ...
    def copy(self) -> MultiDict[_T]: ...
    def __getitem__(self, k: _S) -> _T: ...
    def __setitem__(self, k: _S, v: _T) -> None: ...
    def __delitem__(self, v: _S) -> None: ...
    def __iter__(self) -> Iterator[_S]: ...
    def __len__(self) -> int: ...
    @overload
    def getall(self, key: _S) -> list[_T]: ...
    @overload
    def getall(self, key: _S, default: _D) -> list[_T] | _D: ...
    @overload
    def getone(self, key: _S) -> _T: ...
    @overload
    def getone(self, key: _S, default: _D) -> _T | _D: ...
    def add(self, key: _S, value: _T) -> None: ...
    def extend(self, arg: _Arg[_T] = ..., **kwargs: _T) -> None: ...
    @overload
    def popone(self, key: _S) -> _T: ...
    @overload
    def popone(self, key: _S, default: _D) -> _T | _D: ...
    @overload
    def popall(self, key: _S) -> list[_T]: ...
    @overload
    def popall(self, key: _S, default: _D) -> list[_T] | _D: ...

class CIMultiDict(MutableMultiMapping[_T], Generic[_T]):
    def __init__(self, arg: _Arg[_T] = ..., **kwargs: _T) -> None: ...
    def copy(self) -> CIMultiDict[_T]: ...
    def __getitem__(self, k: _S) -> _T: ...
    def __setitem__(self, k: _S, v: _T) -> None: ...
    def __delitem__(self, v: _S) -> None: ...
    def __iter__(self) -> Iterator[_S]: ...
    def __len__(self) -> int: ...
    @overload
    def getall(self, key: _S) -> list[_T]: ...
    @overload
    def getall(self, key: _S, default: _D) -> list[_T] | _D: ...
    @overload
    def getone(self, key: _S) -> _T: ...
    @overload
    def getone(self, key: _S, default: _D) -> _T | _D: ...
    def add(self, key: _S, value: _T) -> None: ...
    def extend(self, arg: _Arg[_T] = ..., **kwargs: _T) -> None: ...
    @overload
    def popone(self, key: _S) -> _T: ...
    @overload
    def popone(self, key: _S, default: _D) -> _T | _D: ...
    @overload
    def popall(self, key: _S) -> list[_T]: ...
    @overload
    def popall(self, key: _S, default: _D) -> list[_T] | _D: ...

class MultiDictProxy(MultiMapping[_T], Generic[_T]):
    def __init__(
        self, arg: MultiMapping[_T] | MutableMultiMapping[_T]
    ) -> None: ...
    def copy(self) -> MultiDict[_T]: ...
    def __getitem__(self, k: _S) -> _T: ...
    def __iter__(self) -> Iterator[_S]: ...
    def __len__(self) -> int: ...
    @overload
    def getall(self, key: _S) -> list[_T]: ...
    @overload
    def getall(self, key: _S, default: _D) -> list[_T] | _D: ...
    @overload
    def getone(self, key: _S) -> _T: ...
    @overload
    def getone(self, key: _S, default: _D) -> _T | _D: ...

class CIMultiDictProxy(MultiMapping[_T], Generic[_T]):
    def __init__(
        self, arg: MultiMapping[_T] | MutableMultiMapping[_T]
    ) -> None: ...
    def __getitem__(self, k: _S) -> _T: ...
    def __iter__(self) -> Iterator[_S]: ...
    def __len__(self) -> int: ...
    @overload
    def getall(self, key: _S) -> list[_T]: ...
    @overload
    def getall(self, key: _S, default: _D) -> list[_T] | _D: ...
    @overload
    def getone(self, key: _S) -> _T: ...
    @overload
    def getone(self, key: _S, default: _D) -> _T | _D: ...
    def copy(self) -> CIMultiDict[_T]: ...

def getversion(
    md: MultiDict[_T] | CIMultiDict[_T] | MultiDictProxy[_T] | CIMultiDictProxy[_T]
) -> int: ...



import sys
import types
from array import array
from collections import abc

from ._abc import MultiMapping, MutableMultiMapping

_marker = object()

if sys.version_info >= (3, 9):
    GenericAlias = types.GenericAlias
else:
    def GenericAlias(cls):
        return cls


class istr(str):

    """Case insensitive str."""

    __is_istr__ = True


upstr = istr  # for relaxing backward compatibility problems


def getversion(md):
    if not isinstance(md, _Base):
        raise TypeError("Parameter should be multidict or proxy")
    return md._impl._version


_version = array("Q", [0])


class _Impl:
    __slots__ = ("_items", "_version")

    def __init__(self):
        self._items = []
        self.incr_version()

    def incr_version(self):
        global _version
        v = _version
        v[0] += 1
        self._version = v[0]

    if sys.implementation.name != "pypy":

        def __sizeof__(self):
            return object.__sizeof__(self) + sys.getsizeof(self._items)


class _Base:
    def _title(self, key):
        return key

    def getall(self, key, default=_marker):
        """Return a list of all values matching the key."""
        identity = self._title(key)
        res = [v for i, k, v in self._impl._items if i == identity]
        if res:
            return res
        if not res and default is not _marker:
            return default
        raise KeyError("Key not found: %r" % key)

    def getone(self, key, default=_marker):
        """Get first value matching the key.

        Raises KeyError if the key is not found and no default is provided.
        """
        identity = self._title(key)
        for i, k, v in self._impl._items:
            if i == identity:
                return v
        if default is not _marker:
            return default
        raise KeyError("Key not found: %r" % key)

    # Mapping interface #

    def __getitem__(self, key):
        return self.getone(key)

    def get(self, key, default=None):
        """Get first value matching the key.

        If the key is not found, returns the default (or None if no default is provided)
        """
        return self.getone(key, default)

    def __iter__(self):
        return iter(self.keys())

    def __len__(self):
        return len(self._impl._items)

    def keys(self):
        """Return a new view of the dictionary's keys."""
        return _KeysView(self._impl)

    def items(self):
        """Return a new view of the dictionary's items *(key, value) pairs)."""
        return _ItemsView(self._impl)

    def values(self):
        """Return a new view of the dictionary's values."""
        return _ValuesView(self._impl)

    def __eq__(self, other):
        if not isinstance(other, abc.Mapping):
            return NotImplemented
        if isinstance(other, _Base):
            lft = self._impl._items
            rht = other._impl._items
            if len(lft) != len(rht):
                return False
            for (i1, k2, v1), (i2, k2, v2) in zip(lft, rht):
                if i1 != i2 or v1 != v2:
                    return False
            return True
        if len(self._impl._items) != len(other):
            return False
        for k, v in self.items():
            nv = other.get(k, _marker)
            if v != nv:
                return False
        return True

    def __contains__(self, key):
        identity = self._title(key)
        for i, k, v in self._impl._items:
            if i == identity:
                return True
        return False

    def __repr__(self):
        body = ", ".join("'{}': {!r}".format(k, v) for k, v in self.items())
        return "<{}({})>".format(self.__class__.__name__, body)

    __class_getitem__ = classmethod(GenericAlias)


class MultiDictProxy(_Base, MultiMapping):
    """Read-only proxy for MultiDict instance."""

    def __init__(self, arg):
        if not isinstance(arg, (MultiDict, MultiDictProxy)):
            raise TypeError(
                "ctor requires MultiDict or MultiDictProxy instance"
                ", not {}".format(type(arg))
            )

        self._impl = arg._impl

    def __reduce__(self):
        raise TypeError("can't pickle {} objects".format(self.__class__.__name__))

    def copy(self):
        """Return a copy of itself."""
        return MultiDict(self.items())


class CIMultiDictProxy(MultiDictProxy):
    """Read-only proxy for CIMultiDict instance."""

    def __init__(self, arg):
        if not isinstance(arg, (CIMultiDict, CIMultiDictProxy)):
            raise TypeError(
                "ctor requires CIMultiDict or CIMultiDictProxy instance"
                ", not {}".format(type(arg))
            )

        self._impl = arg._impl

    def _title(self, key):
        return key.title()

    def copy(self):
        """Return a copy of itself."""
        return CIMultiDict(self.items())


class MultiDict(_Base, MutableMultiMapping):
    """Dictionary with the support for duplicate keys."""

    def __init__(self, *args, **kwargs):
        self._impl = _Impl()

        self._extend(args, kwargs, self.__class__.__name__, self._extend_items)

    if sys.implementation.name != "pypy":

        def __sizeof__(self):
            return object.__sizeof__(self) + sys.getsizeof(self._impl)

    def __reduce__(self):
        return (self.__class__, (list(self.items()),))

    def _title(self, key):
        return key

    def _key(self, key):
        if isinstance(key, str):
            return key
        else:
            raise TypeError(
                "MultiDict keys should be either str " "or subclasses of str"
            )

    def add(self, key, value):
        identity = self._title(key)
        self._impl._items.append((identity, self._key(key), value))
        self._impl.incr_version()

    def copy(self):
        """Return a copy of itself."""
        cls = self.__class__
        return cls(self.items())

    __copy__ = copy

    def extend(self, *args, **kwargs):
        """Extend current MultiDict with more values.

        This method must be used instead of update.
        """
        self._extend(args, kwargs, "extend", self._extend_items)

    def _extend(self, args, kwargs, name, method):
        if len(args) > 1:
            raise TypeError(
                "{} takes at most 1 positional argument"
                " ({} given)".format(name, len(args))
            )
        if args:
            arg = args[0]
            if isinstance(args[0], (MultiDict, MultiDictProxy)) and not kwargs:
                items = arg._impl._items
            else:
                if hasattr(arg, "items"):
                    arg = arg.items()
                if kwargs:
                    arg = list(arg)
                    arg.extend(list(kwargs.items()))
                items = []
                for item in arg:
                    if not len(item) == 2:
                        raise TypeError(
                            "{} takes either dict or list of (key, value) "
                            "tuples".format(name)
                        )
                    items.append((self._title(item[0]), self._key(item[0]), item[1]))

            method(items)
        else:
            method(
                [
                    (self._title(key), self._key(key), value)
                    for key, value in kwargs.items()
                ]
            )

    def _extend_items(self, items):
        for identity, key, value in items:
            self.add(key, value)

    def clear(self):
        """Remove all items from MultiDict."""
        self._impl._items.clear()
        self._impl.incr_version()

    # Mapping interface #

    def __setitem__(self, key, value):
        self._replace(key, value)

    def __delitem__(self, key):
        identity = self._title(key)
        items = self._impl._items
        found = False
        for i in range(len(items) - 1, -1, -1):
            if items[i][0] == identity:
                del items[i]
                found = True
        if not found:
            raise KeyError(key)
        else:
            self._impl.incr_version()

    def setdefault(self, key, default=None):
        """Return value for key, set value to default if key is not present."""
        identity = self._title(key)
        for i, k, v in self._impl._items:
            if i == identity:
                return v
        self.add(key, default)
        return default

    def popone(self, key, default=_marker):
        """Remove specified key and return the corresponding value.

        If key is not found, d is returned if given, otherwise
        KeyError is raised.

        """
        identity = self._title(key)
        for i in range(len(self._impl._items)):
            if self._impl._items[i][0] == identity:
                value = self._impl._items[i][2]
                del self._impl._items[i]
                self._impl.incr_version()
                return value
        if default is _marker:
            raise KeyError(key)
        else:
            return default

    pop = popone  # type: ignore

    def popall(self, key, default=_marker):
        """Remove all occurrences of key and return the list of corresponding
        values.

        If key is not found, default is returned if given, otherwise
        KeyError is raised.

        """
        found = False
        identity = self._title(key)
        ret = []
        for i in range(len(self._impl._items) - 1, -1, -1):
            item = self._impl._items[i]
            if item[0] == identity:
                ret.append(item[2])
                del self._impl._items[i]
                self._impl.incr_version()
                found = True
        if not found:
            if default is _marker:
                raise KeyError(key)
            else:
                return default
        else:
            ret.reverse()
            return ret

    def popitem(self):
        """Remove and return an arbitrary (key, value) pair."""
        if self._impl._items:
            i = self._impl._items.pop(0)
            self._impl.incr_version()
            return i[1], i[2]
        else:
            raise KeyError("empty multidict")

    def update(self, *args, **kwargs):
        """Update the dictionary from *other*, overwriting existing keys."""
        self._extend(args, kwargs, "update", self._update_items)

    def _update_items(self, items):
        if not items:
            return
        used_keys = {}
        for identity, key, value in items:
            start = used_keys.get(identity, 0)
            for i in range(start, len(self._impl._items)):
                item = self._impl._items[i]
                if item[0] == identity:
                    used_keys[identity] = i + 1
                    self._impl._items[i] = (identity, key, value)
                    break
            else:
                self._impl._items.append((identity, key, value))
                used_keys[identity] = len(self._impl._items)

        # drop tails
        i = 0
        while i < len(self._impl._items):
            item = self._impl._items[i]
            identity = item[0]
            pos = used_keys.get(identity)
            if pos is None:
                i += 1
                continue
            if i >= pos:
                del self._impl._items[i]
            else:
                i += 1

        self._impl.incr_version()

    def _replace(self, key, value):
        key = self._key(key)
        identity = self._title(key)
        items = self._impl._items

        for i in range(len(items)):
            item = items[i]
            if item[0] == identity:
                items[i] = (identity, key, value)
                # i points to last found item
                rgt = i
                self._impl.incr_version()
                break
        else:
            self._impl._items.append((identity, key, value))
            self._impl.incr_version()
            return

        # remove all tail items
        i = rgt + 1
        while i < len(items):
            item = items[i]
            if item[0] == identity:
                del items[i]
            else:
                i += 1


class CIMultiDict(MultiDict):
    """Dictionary with the support for duplicate case-insensitive keys."""

    def _title(self, key):
        return key.title()


class _Iter:
    __slots__ = ("_size", "_iter")

    def __init__(self, size, iterator):
        self._size = size
        self._iter = iterator

    def __iter__(self):
        return self

    def __next__(self):
        return next(self._iter)

    def __length_hint__(self):
        return self._size


class _ViewBase:
    def __init__(self, impl):
        self._impl = impl

    def __len__(self):
        return len(self._impl._items)


class _ItemsView(_ViewBase, abc.ItemsView):
    def __contains__(self, item):
        assert isinstance(item, tuple) or isinstance(item, list)
        assert len(item) == 2
        for i, k, v in self._impl._items:
            if item[0] == k and item[1] == v:
                return True
        return False

    def __iter__(self):
        return _Iter(len(self), self._iter(self._impl._version))

    def _iter(self, version):
        for i, k, v in self._impl._items:
            if version != self._impl._version:
                raise RuntimeError("Dictionary changed during iteration")
            yield k, v

    def __repr__(self):
        lst = []
        for item in self._impl._items:
            lst.append("{!r}: {!r}".format(item[1], item[2]))
        body = ", ".join(lst)
        return "{}({})".format(self.__class__.__name__, body)


class _ValuesView(_ViewBase, abc.ValuesView):
    def __contains__(self, value):
        for item in self._impl._items:
            if item[2] == value:
                return True
        return False

    def __iter__(self):
        return _Iter(len(self), self._iter(self._impl._version))

    def _iter(self, version):
        for item in self._impl._items:
            if version != self._impl._version:
                raise RuntimeError("Dictionary changed during iteration")
            yield item[2]

    def __repr__(self):
        lst = []
        for item in self._impl._items:
            lst.append("{!r}".format(item[2]))
        body = ", ".join(lst)
        return "{}({})".format(self.__class__.__name__, body)


class _KeysView(_ViewBase, abc.KeysView):
    def __contains__(self, key):
        for item in self._impl._items:
            if item[1] == key:
                return True
        return False

    def __iter__(self):
        return _Iter(len(self), self._iter(self._impl._version))

    def _iter(self, version):
        for item in self._impl._items:
            if version != self._impl._version:
                raise RuntimeError("Dictionary changed during iteration")
            yield item[1]

    def __repr__(self):
        lst = []
        for item in self._impl._items:
            lst.append("{!r}".format(item[1]))
        body = ", ".join(lst)
        return "{}({})".format(self.__class__.__name__, body)



# This file doesn't use test parametrization because mypy doesn't nothing about it.
# Concrete types are required

import multidict


def test_classes_not_abstract() -> None:
    d1: multidict.MultiDict[str] = multidict.MultiDict({"a": "b"})
    d2: multidict.CIMultiDict[str] = multidict.CIMultiDict({"a": "b"})

    multidict.MultiDictProxy(d1)
    multidict.CIMultiDictProxy(d2)


def test_getitem() -> None:
    d1: multidict.MultiDict[str] = multidict.MultiDict({"a": "b"})
    d2: multidict.CIMultiDict[str] = multidict.CIMultiDict({"a": "b"})

    d3 = multidict.MultiDictProxy(d1)
    d4 = multidict.CIMultiDictProxy(d2)

    key = multidict.istr("a")

    assert d1["a"] == "b"
    assert d2["a"] == "b"
    assert d3["a"] == "b"
    assert d4["a"] == "b"

    assert d1[key] == "b"
    assert d2[key] == "b"
    assert d3[key] == "b"
    assert d4[key] == "b"


def test_get() -> None:
    d1: multidict.MultiDict[str] = multidict.MultiDict({"a": "b"})
    d2: multidict.CIMultiDict[str] = multidict.CIMultiDict({"a": "b"})

    d3 = multidict.MultiDictProxy(d1)
    d4 = multidict.CIMultiDictProxy(d2)

    key = multidict.istr("a")

    assert d1.get("a") == "b"
    assert d2.get("a") == "b"
    assert d3.get("a") == "b"
    assert d4.get("a") == "b"

    assert d1.get(key) == "b"
    assert d2.get(key) == "b"
    assert d3.get(key) == "b"
    assert d4.get(key) == "b"


def test_get_default() -> None:
    d1: multidict.MultiDict[str] = multidict.MultiDict({"a": "b"})
    d2: multidict.CIMultiDict[str] = multidict.CIMultiDict({"a": "b"})

    d3 = multidict.MultiDictProxy(d1)
    d4 = multidict.CIMultiDictProxy(d2)

    key = multidict.istr("b")

    assert d1.get("b", "d") == "d"
    assert d2.get("b", "d") == "d"
    assert d3.get("b", "d") == "d"
    assert d4.get("b", "d") == "d"

    assert d1.get(key, "d") == "d"
    assert d2.get(key, "d") == "d"
    assert d3.get(key, "d") == "d"
    assert d4.get(key, "d") == "d"


def test_getone() -> None:
    d1: multidict.MultiDict[str] = multidict.MultiDict({"a": "b"})
    d2: multidict.CIMultiDict[str] = multidict.CIMultiDict({"a": "b"})

    d3 = multidict.MultiDictProxy(d1)
    d4 = multidict.CIMultiDictProxy(d2)

    key = multidict.istr("a")

    assert d1.getone("a") == "b"
    assert d2.getone("a") == "b"
    assert d3.getone("a") == "b"
    assert d4.getone("a") == "b"

    assert d1.getone(key) == "b"
    assert d2.getone(key) == "b"
    assert d3.getone(key) == "b"
    assert d4.getone(key) == "b"


def test_getone_default() -> None:
    d1: multidict.MultiDict[str] = multidict.MultiDict({"a": "b"})
    d2: multidict.CIMultiDict[str] = multidict.CIMultiDict({"a": "b"})

    d3 = multidict.MultiDictProxy(d1)
    d4 = multidict.CIMultiDictProxy(d2)

    key = multidict.istr("b")

    assert d1.getone("b", 1) == 1
    assert d2.getone("b", 1) == 1
    assert d3.getone("b", 1) == 1
    assert d4.getone("b", 1) == 1

    assert d1.getone(key, 1) == 1
    assert d2.getone(key, 1) == 1
    assert d3.getone(key, 1) == 1
    assert d4.getone(key, 1) == 1


def test_getall() -> None:
    d1: multidict.MultiDict[str] = multidict.MultiDict({"a": "b"})
    d2: multidict.CIMultiDict[str] = multidict.CIMultiDict({"a": "b"})

    d3 = multidict.MultiDictProxy(d1)
    d4 = multidict.CIMultiDictProxy(d2)

    key = multidict.istr("a")

    assert d1.getall("a") == ["b"]
    assert d2.getall("a") == ["b"]
    assert d3.getall("a") == ["b"]
    assert d4.getall("a") == ["b"]

    assert d1.getall(key) == ["b"]
    assert d2.getall(key) == ["b"]
    assert d3.getall(key) == ["b"]
    assert d4.getall(key) == ["b"]


def test_getall_default() -> None:
    d1: multidict.MultiDict[str] = multidict.MultiDict({"a": "b"})
    d2: multidict.CIMultiDict[str] = multidict.CIMultiDict({"a": "b"})

    d3 = multidict.MultiDictProxy(d1)
    d4 = multidict.CIMultiDictProxy(d2)

    key = multidict.istr("b")

    assert d1.getall("b", 1) == 1
    assert d2.getall("b", 1) == 1
    assert d3.getall("b", 1) == 1
    assert d4.getall("b", 1) == 1

    assert d1.getall(key, 1) == 1
    assert d2.getall(key, 1) == 1
    assert d3.getall(key, 1) == 1
    assert d4.getall(key, 1) == 1


def test_copy() -> None:
    d1: multidict.MultiDict[str] = multidict.MultiDict({"a": "b"})
    d2: multidict.CIMultiDict[str] = multidict.CIMultiDict({"a": "b"})

    d3 = multidict.MultiDictProxy(d1)
    d4 = multidict.CIMultiDictProxy(d2)

    assert d1.copy() == d1
    assert d2.copy() == d2
    assert d3.copy() == d1
    assert d4.copy() == d2


def test_iter() -> None:
    d1: multidict.MultiDict[str] = multidict.MultiDict({"a": "b"})
    d2: multidict.CIMultiDict[str] = multidict.CIMultiDict({"a": "b"})

    d3 = multidict.MultiDictProxy(d1)
    d4 = multidict.CIMultiDictProxy(d2)

    for i in d1:
        i.lower()  # str-specific class
    for i in d2:
        i.lower()  # str-specific class
    for i in d3:
        i.lower()  # str-specific class
    for i in d4:
        i.lower()  # str-specific class


def test_setitem() -> None:
    d1: multidict.MultiDict[str] = multidict.MultiDict({"a": "b"})
    d2: multidict.CIMultiDict[str] = multidict.CIMultiDict({"a": "b"})

    key = multidict.istr("a")

    d1["a"] = "b"
    d2["a"] = "b"

    d1[key] = "b"
    d2[key] = "b"

def test_delitem() -> None:
    d1: multidict.MultiDict[str] = multidict.MultiDict({"a": "b"})
    d2: multidict.CIMultiDict[str] = multidict.CIMultiDict({"a": "b"})

    del d1["a"]
    del d2["a"]

    key = multidict.istr("a")

    d3: multidict.MultiDict[str] = multidict.MultiDict({"a": "b"})
    d4: multidict.CIMultiDict[str] = multidict.CIMultiDict({"a": "b"})

    del d3[key]
    del d4[key]


def test_additem() -> None:
    d1: multidict.MultiDict[str] = multidict.MultiDict({"a": "b"})
    d2: multidict.CIMultiDict[str] = multidict.CIMultiDict({"a": "b"})

    key = multidict.istr("a")

    d1.add("a", "b")
    d2.add("a", "b")

    d1.add(key, "b")
    d2.add(key, "b")


def test_extend_mapping() -> None:
    d1: multidict.MultiDict[str] = multidict.MultiDict({"a": "b"})
    d2: multidict.CIMultiDict[str] = multidict.CIMultiDict({"a": "b"})

    key = multidict.istr("a")

    d1.extend({"a": "b"})
    d2.extend({"a": "b"})

    d1.extend({key: "b"})
    d2.extend({key: "b"})


def test_update_mapping() -> None:
    d1: multidict.MultiDict[str] = multidict.MultiDict({"a": "b"})
    d2: multidict.CIMultiDict[str] = multidict.CIMultiDict({"a": "b"})

    key = multidict.istr("a")

    d1.update({"a": "b"})
    d2.update({"a": "b"})

    d1.update({key: "b"})
    d2.update({key: "b"})

def test_popone() -> None:
    d1: multidict.MultiDict[str] = multidict.MultiDict({"a": "b"})
    d2: multidict.CIMultiDict[str] = multidict.CIMultiDict({"a": "b"})

    assert d1.popone("a") == "b"
    assert d2.popone("a") == "b"

    key = multidict.istr("a")
    d1 = multidict.MultiDict({"a": "b"})
    d2 = multidict.CIMultiDict({"a": "b"})

    assert d1.popone(key) == "b"
    assert d2.popone(key) == "b"


def test_popall() -> None:
    d1: multidict.MultiDict[str] = multidict.MultiDict({"a": "b"})
    d2: multidict.CIMultiDict[str] = multidict.CIMultiDict({"a": "b"})

    assert d1.popall("a") == ["b"]
    assert d2.popall("a") == ["b"]

    key = multidict.istr("a")
    d1 = multidict.MultiDict({"a": "b"})
    d2 = multidict.CIMultiDict({"a": "b"})

    assert d1.popall(key) == ["b"]
    assert d2.popall(key) == ["b"]



import gc
import operator
import sys
import weakref
from collections import deque
from collections.abc import Mapping
from functools import reduce
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    Iterator,
    List,
    Mapping,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
)

import pytest

import multidict
from multidict import CIMultiDict, CIMultiDictProxy, MultiDict, MultiDictProxy

_MultiDictClasses = Union[Type[MultiDict[str]], Type[CIMultiDict[str]]]


def chained_callable(
    module: object, callables: Union[str, Iterable[str]]
) -> Callable[..., Any]:
    """
    Returns callable that will get and call all given objects in module in
    exact order. If `names` is a single object's name function will return
    object itself.

    Will treat `names` of type `str` as a list of single element.
    """
    callables = (callables,) if isinstance(callables, str) else callables
    _callable, *rest = (getattr(module, name) for name in callables)

    def chained_call(*args: object, **kwargs: object) -> Any:
        return reduce(lambda res, c: c(res), rest, _callable(*args, **kwargs))

    return chained_call if len(rest) > 0 else _callable  # type: ignore[no-any-return]


@pytest.fixture(scope="function")
def cls(request: Any, _multidict: Any) -> Any:
    return chained_callable(_multidict, request.param)


@pytest.fixture(scope="function")
def classes(request: Any, _multidict: Any) -> Any:
    return tuple(chained_callable(_multidict, n) for n in request.param)


@pytest.mark.parametrize("cls", ["MultiDict", "CIMultiDict"], indirect=True)
def test_exposed_names(
    cls: Union[Type[MultiDict[object]], Type[CIMultiDict[object]]]
) -> None:
    name = cls.__name__

    while name.startswith("_"):
        name = name[1:]

    assert name in multidict.__all__  # type: ignore[attr-defined]


@pytest.mark.parametrize(
    "cls, key_cls",
    [("MultiDict", str), (("MultiDict", "MultiDictProxy"), str)],
    indirect=["cls"],
)
def test__iter__types(
    cls: Type[MultiDict[Union[str, int]]], key_cls: Type[object]
) -> None:
    d = cls([("key", "one"), ("key2", "two"), ("key", 3)])
    for i in d:
        assert type(i) is key_cls, (type(i), key_cls)


_ClsPair = TypeVar(
    "_ClsPair",
    Tuple[Type[MultiDict[str]], Type[MultiDictProxy[str]]],
    Tuple[Type[CIMultiDict[str]], Type[CIMultiDictProxy[str]]],
)


@pytest.mark.parametrize(
    "classes",
    [("MultiDict", "MultiDictProxy"), ("CIMultiDict", "CIMultiDictProxy")],
    indirect=True,
)
def test_proxy_copy(classes: _ClsPair) -> None:
    dict_cls, proxy_cls = classes
    d1 = dict_cls(key="value", a="b")
    p1 = proxy_cls(d1)

    d2 = p1.copy()
    assert d1 == d2
    assert d1 is not d2


@pytest.mark.parametrize(
    "cls",
    ["MultiDict", "CIMultiDict", "MultiDictProxy", "CIMultiDictProxy"],
    indirect=True,
)
def test_subclassing(cls: Any) -> None:
    class MyClass(cls):  # type: ignore[valid-type,misc]
        pass


class BaseMultiDictTest:
    def test_instantiate__empty(self, cls: _MultiDictClasses) -> None:
        d = cls()
        empty: Mapping[str, str] = {}
        assert d == empty
        assert len(d) == 0
        assert list(d.keys()) == []
        assert list(d.values()) == []
        assert list(d.items()) == []

        assert cls() != list()  # type: ignore[comparison-overlap]
        with pytest.raises(TypeError, match=r"(2 given)"):
            cls(("key1", "value1"), ("key2", "value2"))  # type: ignore[arg-type,call-arg]  # noqa: E501

    @pytest.mark.parametrize("arg0", [[("key", "value1")], {"key": "value1"}])
    def test_instantiate__from_arg0(
        self,
        cls: _MultiDictClasses,
        arg0: Union[List[Tuple[str, str]], Dict[str, str]],
    ) -> None:
        d = cls(arg0)

        assert d == {"key": "value1"}
        assert len(d) == 1
        assert list(d.keys()) == ["key"]
        assert list(d.values()) == ["value1"]
        assert list(d.items()) == [("key", "value1")]

    def test_instantiate__with_kwargs(self, cls: _MultiDictClasses) -> None:
        d = cls([("key", "value1")], key2="value2")

        assert d == {"key": "value1", "key2": "value2"}
        assert len(d) == 2
        assert sorted(d.keys()) == ["key", "key2"]
        assert sorted(d.values()) == ["value1", "value2"]
        assert sorted(d.items()) == [("key", "value1"), ("key2", "value2")]

    def test_instantiate__from_generator(
        self, cls: Union[Type[MultiDict[int]], Type[CIMultiDict[int]]]
    ) -> None:
        d = cls((str(i), i) for i in range(2))

        assert d == {"0": 0, "1": 1}
        assert len(d) == 2
        assert sorted(d.keys()) == ["0", "1"]
        assert sorted(d.values()) == [0, 1]
        assert sorted(d.items()) == [("0", 0), ("1", 1)]

    def test_instantiate__from_list_of_lists(self, cls: _MultiDictClasses) -> None:
        # Should work at runtime, but won't type check.
        d = cls([["key", "value1"]])  # type: ignore[list-item]
        assert d == {"key": "value1"}

    def test_instantiate__from_list_of_custom_pairs(
        self, cls: _MultiDictClasses
    ) -> None:
        class Pair:
            def __len__(self) -> int:
                return 2

            def __getitem__(self, pos: int) -> str:
                if pos == 0:
                    return "key"
                elif pos == 1:
                    return "value1"
                else:
                    raise IndexError

        # Works at runtime, but won't type check.
        d = cls([Pair()])  # type: ignore[list-item]
        assert d == {"key": "value1"}

    def test_getone(self, cls: _MultiDictClasses) -> None:
        d = cls([("key", "value1")], key="value2")

        assert d.getone("key") == "value1"
        assert d.get("key") == "value1"
        assert d["key"] == "value1"

        with pytest.raises(KeyError, match="key2"):
            d["key2"]
        with pytest.raises(KeyError, match="key2"):
            d.getone("key2")

        assert d.getone("key2", "default") == "default"

    def test__iter__(
        self,
        cls: Union[Type[MultiDict[Union[str, int]]], Type[CIMultiDict[Union[str, int]]]]
    ) -> None:
        d = cls([("key", "one"), ("key2", "two"), ("key", 3)])
        assert list(d) == ["key", "key2", "key"]

    def test_keys__contains(
        self,
        cls: Union[Type[MultiDict[Union[str, int]]], Type[CIMultiDict[Union[str, int]]]]
    ) -> None:
        d = cls([("key", "one"), ("key2", "two"), ("key", 3)])

        assert list(d.keys()) == ["key", "key2", "key"]

        assert "key" in d.keys()
        assert "key2" in d.keys()

        assert "foo" not in d.keys()

    def test_values__contains(
        self,
        cls: Union[Type[MultiDict[Union[str, int]]], Type[CIMultiDict[Union[str, int]]]]
    ) -> None:
        d = cls([("key", "one"), ("key", "two"), ("key", 3)])

        assert list(d.values()) == ["one", "two", 3]

        assert "one" in d.values()
        assert "two" in d.values()
        assert 3 in d.values()

        assert "foo" not in d.values()

    def test_items__contains(
        self,
        cls: Union[Type[MultiDict[Union[str, int]]], Type[CIMultiDict[Union[str, int]]]]
    ) -> None:
        d = cls([("key", "one"), ("key", "two"), ("key", 3)])

        assert list(d.items()) == [("key", "one"), ("key", "two"), ("key", 3)]

        assert ("key", "one") in d.items()
        assert ("key", "two") in d.items()
        assert ("key", 3) in d.items()

        assert ("foo", "bar") not in d.items()

    def test_cannot_create_from_unaccepted(self, cls: _MultiDictClasses) -> None:
        with pytest.raises(TypeError):
            cls([(1, 2, 3)])  # type: ignore[list-item]

    def test_keys_is_set_less(self, cls: _MultiDictClasses) -> None:
        d = cls([("key", "value1")])

        assert d.keys() < {"key", "key2"}

    def test_keys_is_set_less_equal(self, cls: _MultiDictClasses) -> None:
        d = cls([("key", "value1")])

        assert d.keys() <= {"key"}

    def test_keys_is_set_equal(self, cls: _MultiDictClasses) -> None:
        d = cls([("key", "value1")])

        assert d.keys() == {"key"}

    def test_keys_is_set_greater(self, cls: _MultiDictClasses) -> None:
        d = cls([("key", "value1")])

        assert {"key", "key2"} > d.keys()

    def test_keys_is_set_greater_equal(self, cls: _MultiDictClasses) -> None:
        d = cls([("key", "value1")])

        assert {"key"} >= d.keys()

    def test_keys_is_set_not_equal(self, cls: _MultiDictClasses) -> None:
        d = cls([("key", "value1")])

        assert d.keys() != {"key2"}

    def test_eq(self, cls: _MultiDictClasses) -> None:
        d = cls([("key", "value1")])

        assert {"key": "value1"} == d

    def test_eq2(self, cls: _MultiDictClasses) -> None:
        d1 = cls([("key", "value1")])
        d2 = cls([("key2", "value1")])

        assert d1 != d2

    def test_eq3(self, cls: _MultiDictClasses) -> None:
        d1 = cls([("key", "value1")])
        d2 = cls()

        assert d1 != d2

    def test_eq_other_mapping_contains_more_keys(self, cls: _MultiDictClasses) -> None:
        d1 = cls(foo="bar")
        d2 = dict(foo="bar", bar="baz")

        assert d1 != d2

    def test_eq_bad_mapping_len(
        self, cls: Union[Type[MultiDict[int]], Type[CIMultiDict[int]]]
    ) -> None:
        class BadMapping(Mapping[str, int]):
            def __getitem__(self, key: str) -> int:
                return 1

            def __iter__(self) -> Iterator[str]:
                yield "a"

            def __len__(self) -> int:  # type: ignore[return]
                1 / 0

        d1 = cls(a=1)
        d2 = BadMapping()
        with pytest.raises(ZeroDivisionError):
            d1 == d2

    def test_eq_bad_mapping_getitem(
        self,
        cls: Union[Type[MultiDict[int]], Type[CIMultiDict[int]]]
    ) -> None:
        class BadMapping(Mapping[str, int]):
            def __getitem__(self, key: str) -> int:  # type: ignore[return]
                1 / 0

            def __iter__(self) -> Iterator[str]:
                yield "a"

            def __len__(self) -> int:
                return 1

        d1 = cls(a=1)
        d2 = BadMapping()
        with pytest.raises(ZeroDivisionError):
            d1 == d2

    def test_ne(self, cls: _MultiDictClasses) -> None:
        d = cls([("key", "value1")])

        assert d != {"key": "another_value"}

    def test_and(self, cls: _MultiDictClasses) -> None:
        d = cls([("key", "value1")])

        assert {"key"} == d.keys() & {"key", "key2"}

    def test_and2(self, cls: _MultiDictClasses) -> None:
        d = cls([("key", "value1")])

        assert {"key"} == {"key", "key2"} & d.keys()

    def test_or(self, cls: _MultiDictClasses) -> None:
        d = cls([("key", "value1")])

        assert {"key", "key2"} == d.keys() | {"key2"}

    def test_or2(self, cls: _MultiDictClasses) -> None:
        d = cls([("key", "value1")])

        assert {"key", "key2"} == {"key2"} | d.keys()

    def test_sub(self, cls: _MultiDictClasses) -> None:
        d = cls([("key", "value1"), ("key2", "value2")])

        assert {"key"} == d.keys() - {"key2"}

    def test_sub2(self, cls: _MultiDictClasses) -> None:
        d = cls([("key", "value1"), ("key2", "value2")])

        assert {"key3"} == {"key", "key2", "key3"} - d.keys()

    def test_xor(self, cls: _MultiDictClasses) -> None:
        d = cls([("key", "value1"), ("key2", "value2")])

        assert {"key", "key3"} == d.keys() ^ {"key2", "key3"}

    def test_xor2(self, cls: _MultiDictClasses) -> None:
        d = cls([("key", "value1"), ("key2", "value2")])

        assert {"key", "key3"} == {"key2", "key3"} ^ d.keys()

    @pytest.mark.parametrize("_set, expected", [({"key2"}, True), ({"key"}, False)])
    def test_isdisjoint(
        self, cls: _MultiDictClasses, _set: Set[str], expected: bool
    ) -> None:
        d = cls([("key", "value1")])

        assert d.keys().isdisjoint(_set) == expected

    def test_repr_issue_410(self, cls: _MultiDictClasses) -> None:
        d = cls()

        try:
            raise Exception
            pytest.fail("Should never happen")  # pragma: no cover
        except Exception as e:
            repr(d)

            assert sys.exc_info()[1] == e

    @pytest.mark.parametrize(
        "op", [operator.or_, operator.and_, operator.sub, operator.xor]
    )
    @pytest.mark.parametrize("other", [{"other"}])
    def test_op_issue_410(
        self,
        cls: _MultiDictClasses,
        op: Callable[[object, object], object],
        other: Set[str],
    ) -> None:
        d = cls([("key", "value")])

        try:
            raise Exception
            pytest.fail("Should never happen")  # pragma: no cover
        except Exception as e:
            op(d.keys(), other)

            assert sys.exc_info()[1] == e

    def test_weakref(self, cls: _MultiDictClasses) -> None:
        called = False

        def cb(wr: object) -> None:
            nonlocal called
            called = True

        d = cls()
        wr = weakref.ref(d, cb)
        del d
        gc.collect()
        assert called
        del wr

    def test_iter_length_hint_keys(
        self,
        cls: Union[Type[MultiDict[int]], Type[CIMultiDict[int]]]
    ) -> None:
        md = cls(a=1, b=2)
        it = iter(md.keys())
        assert it.__length_hint__() == 2  # type: ignore[attr-defined]

    def test_iter_length_hint_items(
        self,
        cls: Union[Type[MultiDict[int]], Type[CIMultiDict[int]]]
    ) -> None:
        md = cls(a=1, b=2)
        it = iter(md.items())
        assert it.__length_hint__() == 2  # type: ignore[attr-defined]

    def test_iter_length_hint_values(
        self,
        cls: Union[Type[MultiDict[int]], Type[CIMultiDict[int]]]
    ) -> None:
        md = cls(a=1, b=2)
        it = iter(md.values())
        assert it.__length_hint__() == 2  # type: ignore[attr-defined]

    def test_ctor_list_arg_and_kwds(
        self,
        cls: Union[Type[MultiDict[int]], Type[CIMultiDict[int]]]
    ) -> None:
        arg = [("a", 1)]
        obj = cls(arg, b=2)
        assert list(obj.items()) == [("a", 1), ("b", 2)]
        assert arg == [("a", 1)]

    def test_ctor_tuple_arg_and_kwds(
        self,
        cls: Union[Type[MultiDict[int]], Type[CIMultiDict[int]]]
    ) -> None:
        arg = (("a", 1),)
        obj = cls(arg, b=2)
        assert list(obj.items()) == [("a", 1), ("b", 2)]
        assert arg == (("a", 1),)

    def test_ctor_deque_arg_and_kwds(
        self,
        cls: Union[Type[MultiDict[int]], Type[CIMultiDict[int]]]
    ) -> None:
        arg = deque([("a", 1)])
        obj = cls(arg, b=2)
        assert list(obj.items()) == [("a", 1), ("b", 2)]
        assert arg == deque([("a", 1)])


class TestMultiDict(BaseMultiDictTest):
    @pytest.fixture(params=["MultiDict", ("MultiDict", "MultiDictProxy")])
    def cls(self, request: Any, _multidict: Any) -> Any:
        return chained_callable(_multidict, request.param)

    def test__repr__(self, cls: Type[MultiDict[str]]) -> None:
        d = cls()
        _cls = type(d)

        assert str(d) == "<%s()>" % _cls.__name__

        d = cls([("key", "one"), ("key", "two")])

        assert str(d) == "<%s('key': 'one', 'key': 'two')>" % _cls.__name__

    def test_getall(self, cls: Type[MultiDict[str]]) -> None:
        d = cls([("key", "value1")], key="value2")

        assert d != {"key": "value1"}
        assert len(d) == 2

        assert d.getall("key") == ["value1", "value2"]

        with pytest.raises(KeyError, match="some_key"):
            d.getall("some_key")

        default = object()
        assert d.getall("some_key", default) is default

    def test_preserve_stable_ordering(
        self, cls: Type[MultiDict[Union[str, int]]]
    ) -> None:
        d = cls([("a", 1), ("b", "2"), ("a", 3)])
        s = "&".join("{}={}".format(k, v) for k, v in d.items())

        assert s == "a=1&b=2&a=3"

    def test_get(self, cls: Type[MultiDict[int]]) -> None:
        d = cls([("a", 1), ("a", 2)])
        assert d["a"] == 1

    def test_items__repr__(self, cls: Type[MultiDict[str]]) -> None:
        d = cls([("key", "value1")], key="value2")
        expected = "_ItemsView('key': 'value1', 'key': 'value2')"
        assert repr(d.items()) == expected

    def test_keys__repr__(self, cls: Type[MultiDict[str]]) -> None:
        d = cls([("key", "value1")], key="value2")
        assert repr(d.keys()) == "_KeysView('key', 'key')"

    def test_values__repr__(self, cls: Type[MultiDict[str]]) -> None:
        d = cls([("key", "value1")], key="value2")
        assert repr(d.values()) == "_ValuesView('value1', 'value2')"


class TestCIMultiDict(BaseMultiDictTest):
    @pytest.fixture(params=["CIMultiDict", ("CIMultiDict", "CIMultiDictProxy")])
    def cls(self, request: Any, _multidict: Any) -> Any:
        return chained_callable(_multidict, request.param)

    def test_basics(self, cls: Type[CIMultiDict[str]]) -> None:
        d = cls([("KEY", "value1")], KEY="value2")

        assert d.getone("key") == "value1"
        assert d.get("key") == "value1"
        assert d.get("key2", "val") == "val"
        assert d["key"] == "value1"
        assert "key" in d

        with pytest.raises(KeyError, match="key2"):
            d["key2"]
        with pytest.raises(KeyError, match="key2"):
            d.getone("key2")

    def test_getall(self, cls: Type[CIMultiDict[str]]) -> None:
        d = cls([("KEY", "value1")], KEY="value2")

        assert not d == {"KEY": "value1"}
        assert len(d) == 2

        assert d.getall("key") == ["value1", "value2"]

        with pytest.raises(KeyError, match="some_key"):
            d.getall("some_key")

    def test_get(self, cls: Type[CIMultiDict[int]]) -> None:
        d = cls([("A", 1), ("a", 2)])
        assert 1 == d["a"]

    def test__repr__(self, cls: Type[CIMultiDict[str]]) -> None:
        d = cls([("KEY", "value1")], key="value2")
        _cls = type(d)

        expected = "<%s('KEY': 'value1', 'key': 'value2')>" % _cls.__name__
        assert str(d) == expected

    def test_items__repr__(self, cls: Type[CIMultiDict[str]]) -> None:
        d = cls([("KEY", "value1")], key="value2")
        expected = "_ItemsView('KEY': 'value1', 'key': 'value2')"
        assert repr(d.items()) == expected

    def test_keys__repr__(self, cls: Type[CIMultiDict[str]]) -> None:
        d = cls([("KEY", "value1")], key="value2")
        assert repr(d.keys()) == "_KeysView('KEY', 'key')"

    def test_values__repr__(self, cls: Type[CIMultiDict[str]]) -> None:
        d = cls([("KEY", "value1")], key="value2")
        assert repr(d.values()) == "_ValuesView('value1', 'value2')"
